import torch.nn as nn
import torch


class OneRelClassifier(nn.Module):
    
    def __init__(self,
                 input_size: int,
                 output_size: int,
                 tag_size: int,
                 dropout_prob: float):
        super().__init__()
        self.output_size = output_size
        self.tag_size =tag_size
        self.linear1 = nn.Linear(input_size*2, input_size*3)
        self.dropout =  nn.Dropout(dropout_prob)
        self.activation = nn.ReLU()
        self.linear2 = nn.Linear(input_size*3, tag_size*output_size)
        
    def forward(self, x): 
        batch_size, seq_len, hidden_size = x.size()
        # head: [batch_size, seq_len * seq_len, hidden_size] 重复样式为1,1,1, 2,2,2, 3,3,3
        head = x.unsqueeze(2).expand(batch_size, seq_len, seq_len, hidden_size).reshape(batch_size, seq_len*seq_len, hidden_size)
        # tail: [batch_size, seq_len * seq_len, hidden_size] 重复样式为1,2,3, 1,2,3, 1,2,3
        tail = x.repeat(1, seq_len, 1)
        # 两两token的特征拼接在一起
        pairs = torch.cat([head, tail], dim=-1)
        pairs = self.linear1(pairs)
        pairs = self.dropout(pairs)
        pairs = self.activation(pairs)
        scores = self.linear2(pairs).reshape(batch_size, seq_len, seq_len, self.output_size, self.tag_size)
        return scores


from ...utils.make_model import PLMBaseModel, align_token_span
from ...layers.classifier import BiaffineSpanClassifier
from ...layers.dropout import MultiDropout
from ...layers.loss import MultiLabelCategoricalCrossEntropy
from ...metrics.span import SpanF1
import torch


class BiaffineForEntityExtraction(PLMBaseModel):
    def __init__(self,
                 lr: float = 3e-5,
                 hidden_size: int = 128,
                 add_rope: bool = True,
                 threshold: float = 0.0,
                 scheduler: str = 'linear_warmup_step',
                 weight_decay: float = 0.01,
                 **kwargs) -> None:
        super().__init__()
        
        self.plm = self.get_plm_architecture()
        self.classifier = BiaffineSpanClassifier(input_size=self.plm.config.hidden_size,
                                                 hidden_size=hidden_size,
                                                 output_size=len(self.hparams.label2id),
                                                 add_rope=add_rope)
        self.dropout = MultiDropout()
        
        self.criterion = MultiLabelCategoricalCrossEntropy()

        self.train_metric = SpanF1()
        self.val_metric = SpanF1()
        self.test_metric = SpanF1()


    def forward(self, input_ids, token_type_ids, attention_mask):
        x = self.plm(input_ids=input_ids, token_type_ids=token_type_ids, attention_mask=attention_mask).last_hidden_state
        x = self.dropout(x)
        x = self.classifier(x, mask=attention_mask)
        return x
    
    
    def step(self, batch):
        input_ids = batch['input_ids']
        token_type_ids = batch['token_type_ids']
        attention_mask = batch['attention_mask']
        span_ids = batch['label_ids']
        logits = self(input_ids, token_type_ids, attention_mask)
        pred = logits.ge(self.hparams.threshold).float()
        batch_size, ent_type_size = logits.shape[:2]
        y_true = span_ids.reshape(batch_size*ent_type_size, -1)
        y_pred = logits.reshape(batch_size*ent_type_size, -1)
        loss = self.criterion(y_pred, y_true)
        return loss, pred, span_ids

        
    def training_step(self, batch, batch_idx):
        loss, pred, true = self.step(batch)
        self.train_metric(pred, true)
        self.log('train/f1', self.train_metric, on_step=True, on_epoch=True, prog_bar=True)
        return {'loss':loss}
    
    
    def validation_step(self, batch, batch_idx):
        loss, pred, true = self.step(batch)
        self.val_metric(pred, true)
        self.log('val/f1', self.val_metric, on_step=False, on_epoch=True, prog_bar=True)
        return {'loss':loss}
    
    
    def test_step(self, batch, batch_idx):
        loss, pred, true = self.shared_step(batch)
        self.test_metric(pred, true)
        self.log('test/f1', self.test_metric, on_step=False, on_epoch=True, prog_bar=True)
        return {'loss': loss}
    
    
    def configure_optimizers(self):
        no_decay = ['bias', 'LayerNorm.bias', 'LayerNorm.weight']
        grouped_parameters = [
            {'params': [p for n, p in self.plm.named_parameters() if not any(nd in n for nd in no_decay)],
             'lr': self.hparams.lr, 'weight_decay': self.hparams.weight_decay},
            {'params': [p for n, p in self.plm.named_parameters() if any(nd in n for nd in no_decay)],
             'lr': self.hparams.lr, 'weight_decay': 0.0},
            {'params': [p for n, p in self.classifier.named_parameters() if not any(nd in n for nd in no_decay)],
             'lr': self.hparams.lr * 5, 'weight_decay': self.hparams.weight_decay},
            {'params': [p for n, p in self.classifier.named_parameters() if any(nd in n for nd in no_decay)],
             'lr': self.hparams.lr * 5, 'weight_decay': 0.0}
        ]
        optimizer = torch.optim.AdamW(grouped_parameters)
        scheduler_config = self.get_scheduler_config(optimizer, self.hparams.scheduler)
        return [optimizer], [scheduler_config]
    
    
    def predict(self, text: str, device: str='cpu', threshold = None):
        if threshold is None:
            threshold = self.hparams.threshold
        inputs = self.tokenizer(text,
                                add_special_tokens=True,
                                max_length=self.hparams.max_length,
                                truncation=True,
                                return_offsets_mapping=True,
                                return_tensors='pt')
        mapping = inputs.pop('offset_mapping')
        mapping = mapping[0].tolist()
        inputs.to(device)
        logits = self(**inputs)
        spans_ls = torch.nonzero(logits>threshold).tolist()
        spans = []
        for span in spans_ls :
            start = span[2]
            end = span[3]
            char_span = align_token_span((start, end+1), mapping)
            start = char_span[0]
            end = char_span[1]
            span_text = text[start:end]
            spans.append([start, end, self.hparams.id2label[span[1]], span_text])
        return spans


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nlhappy',
 'nlhappy.algorithms',
 'nlhappy.callbacks',
 'nlhappy.components',
 'nlhappy.configs',
 'nlhappy.datamodules',
 'nlhappy.layers',
 'nlhappy.layers.attention',
 'nlhappy.layers.classifier',
 'nlhappy.layers.embedding',
 'nlhappy.metrics',
 'nlhappy.models',
 'nlhappy.models.entity_extraction',
 'nlhappy.models.event_extraction',
 'nlhappy.models.prompt_relation_extraction',
 'nlhappy.models.prompt_span_extraction',
 'nlhappy.models.relation_extraction',
 'nlhappy.models.span_extraction',
 'nlhappy.models.text_classification',
 'nlhappy.models.text_multi_classification',
 'nlhappy.models.text_pair_classification',
 'nlhappy.models.text_pair_regression',
 'nlhappy.models.token_classification',
 'nlhappy.tricks',
 'nlhappy.utils']

package_data = \
{'': ['*'],
 'nlhappy.configs': ['callbacks/*',
                     'datamodule/*',
                     'log_dir/*',
                     'logger/*',
                     'model/*',
                     'trainer/*']}

install_requires = \
['datasets>=2.0.0',
 'googletrans==4.0.0rc1',
 'hydra-colorlog>=1.1.0',
 'hydra-core==1.1',
 'pytorch-lightning==1.7.7',
 'rich>=12.4.3,<13.0.0',
 'torch>=1.11.0',
 'transformers>=4.17.0']

extras_require = \
{':extra == "spacy" or extra == "all"': ['spacy>=3.3.0'],
 ':extra == "wandb" or extra == "all"': ['wandb>=0.12.18']}

entry_points = \
{'console_scripts': ['nlhappy = nlhappy.run:train'],
 'spacy_factories': ['span_classifier = nlhappy.components:get_spancat']}

setup_kwargs = {
    'name': 'nlhappy',
    'version': '2022.10.20',
    'description': '自然语言处理(NLP)',
    'long_description': '\n<div align=\'center\'>\n\n# NLHappy\n<a href="https://pytorch.org/get-started/locally/"><img alt="PyTorch" src="https://img.shields.io/badge/PyTorch-ee4c2c?logo=pytorch&logoColor=white"></a>\n<a href="https://pytorchlightning.ai/"><img alt="Lightning" src="https://img.shields.io/badge/-Lightning-792ee5?logo=pytorchlightning&logoColor=white"></a>\n<a href="https://hydra.cc/"><img alt="Config: Hydra" src="https://img.shields.io/badge/Config-Hydra-89b8cd"></a>\n<a href="https://github.com/ashleve/lightning-hydra-template"><img alt="Template" src="https://img.shields.io/badge/-Lightning--Hydra--Template-017F2F?style=flat&logo=github&labelColor=gray"></a>\n<a href="https://spacy.io/"><img alt="Spacy" src="https://img.shields.io/badge/component-%20Spacy-blue"></a>\n<a href="https://wandb.ai/"><img alt="WanDB" src="https://img.shields.io/badge/Log-WanDB-brightgreen"></a>\n</div>\n<br><br>\n\n## 📌&nbsp;&nbsp; 简介\n\nnlhappy是一款集成了数据处理,模型训练,文本处理流程构建等各种功能的自然语言处理库,并且内置了各种任务的SOTA方案,相信通过nlhappy可以让你更愉悦的做各种nlp任务\n> 它主要的依赖有\n- [pytorch-lightning](https://pytorch-lightning.readthedocs.io/en/latest/): 用于模型的训练\n- [transformers](https://huggingface.co/docs/transformers/index): 获取预训练模型\n- [datasets](https://huggingface.co/docs/datasets/index): 构建数据集\n- [spacy](https://spacy.io/usage): 用于自然语言处理流程和组件构建\n\n\n## 🚀&nbsp;&nbsp; 安装\n<details>\n<summary><b>安装nlhappy</b></summary>\n\n> 推荐先去[pytorch官网](https://pytorch.org/get-started/locally/)安装pytorch和对应cuda\n```bash\n# pip 安装\npip install --upgrade pip\npip install --upgrade nlhappy\n```\n</details>\n\n<details>\n<summary><b>其他可选</b></summary>\n\n> 推荐安装wandb用于可视化训练日志\n- 注册: https://wandb.ai/\n- 获取认证: https://wandb.ai/authorize\n- 登陆:\n```bash\nwandb login\n```\n模型训练开始后去[官网](https://wandb.ai/)查看训练实况\n</details>\n\n\n\n\n## ⚡&nbsp;&nbsp; 快速开始\n\n> 任务示例的数据集获取:[CBLUE官网](https://tianchi.aliyun.com/dataset/dataDetail?dataId=95414)\n\n<details>\n<summary><b>文本分类</b></summary>\n\n> 制作数据集\n```python\nfrom nlhappy.datamodules import TextClassificationDataModule\nfrom nlhappy.utils.make_dataset import DatasetDict, Dataset\nimport srsly\n\n# CBLUE短文本分类数据集CHIP-CTC为例\n\n# 查看数据集格式\nexample = TextClassificationDataModule.show_one_example()\n# example : {\'label\': \'新闻\', \'text\': \'怎么给这个图片添加超级链接呢？\'}\n\n# 制作数据集\ntrain_data = list(srsly.read_json(\'assets/CHIP-CTC/CHIP-CTC_train.json\'))\nval_data = list(srsly.read_json(\'assets/CHIP-CTC/CHIP-CTC_dev.json\'))\ndef convert_to_dataset(data):\n    dataset = {\'text\':[], \'label\':[]}\n    for d in data:\n        dataset[\'text\'].append(d[\'text\'])\n        dataset[\'label\'].append(d[\'label\'])\n    return Dataset.from_dict(dataset)\ntrain_ds = convert_to_dataset(train_data)\nval_ds = convert_to_dataset(val_data)\ndataset_dict = DatasetDict({\'train\':train_ds, \'validation\':val_ds})\n\n# 保存数据集, datasets为nlhappy默认数据集路径\ndataset_dict.save_to_disk(\'./datasets/CHIP-CTC\')\n```\n> 训练模型\n\n- 命令行或bash脚本,方便快捷\n```\nnlhappy \\\ndatamodule=text_classification \\\ndatamodule.dataset=TNEWS \\\n# huffingface的预训练模型都会支持\ndatamodule.plm=hfl/chinese-roberta-wwm-ext \\ \ndatamodule.batch_size=32 \\\nmodel=bert_tc \\\nmodel.lr=3e-5 \\\nseed=1234 \\\ntrainer=default\n# 默认为单gpu 0号显卡训练,可以通过以下方式修改显卡\n# trainer.devices=[1]\n# 单卡半精度训练\n# trainer.precision=16\n# 使用wandb记录日志\n# logger=wandb\n# 多卡训练\n# trainer=ddp trainer.devices=4\n```\n\n> 模型预测\n```python\nfrom nlhappy.models import BertTextClassification\n# 加载训练的checkpoint\nckpt = \'logs/path/***.ckpt\'\nmodel = BertTextClassification.load_from_ckeckpoint(ckpt)\ntext = \'研究开始前30天内，接受过其他临床方案治疗；\'\nscores = model.predict(text=text, device=\'cpu\')\n# 转为onnx模型\nmodel.to_onnx(\'path/tc.onnx\')\nmodel.tokenizer.save_pretrained(\'path/tokenizer\')\n```\n</details>\n\n<details>\n<summary><b>实体抽取</b></summary>\n\nnlhappy支持正常,嵌套和非连续的实体抽取任务,下面以可以解决嵌套任务的模型globalpointer为例\n> 制作数据集\n```python\n# CBLUE实体识别数据集CMeEE\nfrom nlhappy.datamodules import EntityExtractionDataModule\nfrom nlhappy.models.entity_extraction import GlobalPointerForEntityExtraction\nfrom nlhappy.utils.make_dataset import Dataset, DatasetDict\nimport srsly\n\n#查看数据集格式\nexample = EntityExtractionDataModule.show_one_example()\n# example : {"text":"这是一个长颈鹿","entities":[{"indexes":[4,5,6],"label":"动物", "text":"长颈鹿"}]}\n\n# 根据格式制作数据集\ntrain_data = list(srsly.read_json(\'assets/CMeEE/CMeEE_train.json\'))\nval_data = list(srsly.read_json(\'assets/CMeEE/CMeEE_dev.json\'))\ndef convert_to_dataset(data):\n    ds = {\'text\':[],\'entities\':[]}\n    for d in data:\n        ents = []\n        ds[\'text\'].append(d[\'text\'])\n        for e in d[\'entities\']:\n            if len(e[\'entity\'])>0:\n                ent = {}\n                ent[\'text\'] = e[\'entity\']\n                ent[\'indexes\'] = [idx for idx in range(len(e[\'entity\']))]\n                ent[\'text\'] = e[\'entity\']\n                ent[\'label\'] = e[\'type\']\n                ents.append(ent)\n        ds[\'entities\'].append(ents)\n    return Dataset.from_dict(ds) \ntrain_ds = convert_to_dataset(train_data)\nval_ds = convert_to_dataset(val_data)\ndataset_dict = DatasetDict({\'train\':train_ds, \'validation\':val_ds})\ndataset_dict.save_to_disk(\'./datasets/CMeEE\')\n\n\n```\n> 训练模型\n\n- 编写训练脚本\n```bash\nnlhappy \\\ndatamodule=entity_extraction \\\ndatamodule.dataset=CMeEE \\\ndatamodule.plm=hfl/chinese-roberta-wwm-ext \\\ndatamodule.batch_size=16 \\\nmodel=globalpointer \\\nmodel.lr=3e-5 \\\nseed=123\n# 默认trainer为单卡训练\ntrainer=default\n\n# 多卡\n# trainer=ddp \\\n# trainer.devices=4 \\\n# trainer.precision=16 #半精度\n\n# 如果安装了wandb则可以使用wandb\n# logger=wandb \n```\n\n> 模型推理\n\n```python\nfrom nlhappy.models import GlobalPointer\nckpt = \'logs/path/***.ckpt\'\nmodel = GlobalPointer.load_from_ckeckpoint(ckpt)\nents = model.predcit(\'文本\')\n\n# 转为onnx模型,进行后续部署\nmodel.to_onnx(\'path/tc.onnx\')\nmodel.tokenizer.save_pretrained(\'path/tokenizer\')\n```\n</details>\n\n<details>\n<summary><b>关系抽取</b></summary>\nTODO\n</details>\n\n<details>\n<summary><b>事件抽取</b></summary>\nTODO\n</details>\n\n<details>\n<summary><b>摘要</b></summary>\nTODO\n</details>\n\n<details>\n<summary><b>翻译</b></summary>\nTODO\n</details>\n\n\n## 论文复现',
    'author': 'wangmengdi',
    'author_email': '790990241@qq.om',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from ..models import GlobalPointer
from spacy.lang.zh import Chinese
import torch
from spacy.tokens import Doc
import os
from spacy.util import filter_spans
import pickle
from nlhappy.utils.utils import get_logger


log = get_logger()

class SpanClassification:
    '''句子级别span分类spacy pipeline
    - ckpt: 模型保存路径
    '''
    def __init__(self, nlp, name:str, device:str, sentence_level:bool, threshold:float, set_ents: bool):
        self.nlp = nlp
        self.pipe_name = name
        self.set_ents = set_ents
        self.sentence_level = sentence_level
        self.device = torch.device(device)
        self.threshold = threshold
        
        
    def __call__(self, doc: Doc) -> Doc:
        all_spans = []
        if self.sentence_level:
            for sent in doc.sents:
                spans = self.model.predict(sent.text, device=self.device, threshold=self.threshold)
                for span in spans:
                    s = sent.char_span(span[0], span[1], span[2])
                    all_spans.append(s)
            doc.spans['all'] = all_spans
        else:
            spans = self.model.predict(doc.text, device=self.device)
            for span in spans:
                s = doc.char_span(span[0], span[1], span[2])
                all_spans.append(s)
            doc.spans['all'] = all_spans
        if self.set_ents:
            doc.set_ents(filter_spans(doc.spans['all']))
        return doc

    def to_disk(self, path:str, exclude):
        # 复制原来模型参数到新的路径
        if not os.path.exists(path):
            os.mkdir(path=path)
        model = 'sc.pkl'
        model_path = os.path.join(path, model)
        with open(model_path, 'wb') as f:
            pickle.dump(self.model, f)

    def from_disk(self, path:str, exclude):
        model_path = os.path.join(path, 'sc.pkl')
        with open(model_path, 'rb') as f:
            self.model = pickle.load(f)
        self.model.freeze()
        try:
            self.model.to(self.device)
        except:
            log.info(f' to device {self.device} failed')
            
    def init_model(self, model_or_path):
        if isinstance(model_or_path, GlobalPointer):
            self.model = model_or_path
            self.model.freeze()
            self.model.to(self.device)
            
        else:
            self.model= GlobalPointer.load_from_checkpoint(model_or_path)
            self.model.freeze()
            self.model.to(self.device)
            
default_config={'device':'cpu', 'sentence_level':False, 'threshold':0.5,'set_ents':True}

@Chinese.factory('span_classifier',assigns=['doc.spans'],default_config=default_config)
def get_spancat(nlp, name:str, device:str, sentence_level:bool, threshold:float, set_ents: bool):
    """句子级别的文本片段分类"""
    return SpanClassification(nlp=nlp, name=name, device=device, sentence_level=sentence_level, threshold=threshold, set_ents=set_ents)





explain_dict = {
    'PSE': 'Prompt Span Extraction, 基于提示的文本抽取简称',
    'TC': 'text classification, 文本分类简称',
    'EE': 'Event Extraction, 事件抽取简称',
    'ToC': 'Token Classification, 基于token的分类简称',
    'DuUIE': '2022CCKS通用信息抽取数据集, 包含实体抽取、关系抽取、事件抽取和情感抽取四个任务.',
    'TPC': 'text-pair classification, 文本对分类简称',
    'TPS': 'text-pair similarity 文本对相似度的简称',
    'ChnSentiCorp': '经典的句子级情感分类数据集，包含酒店、笔记本电脑和数据相关的网络评论数据，共包含积极、消极两个类别。',
    'JDNER': '京东商城商品标题命名实体识别数据集. 标签全部为脱敏数字, 标签非常不均衡.',
    'TNEWS': '今日头条中文新闻（短文本）分类, 该数据集来自今日头条的新闻版块，共提取了15个类别的新闻，包括旅游，教育，金融，军事等',
    'CMeIE': 'CBLUE排行榜中中文医学文本实体关系抽取任务数据集,文本中含有@,并且语义模糊.',
    'CMeEE': 'CBLUE排行榜中中文医学命名实体识别数据集, 含有嵌套的情况.',
    'LCQMC': '中文问答匹配数据集, 238766训练集、8802验证集和12500测试集',
    'ICD-STS': 'ICD-STS数据集, 包含训练集、验证集和测试集',
    'CHIP-CDN': 'CBLUE排行榜中临床术语标准化任务的数据集, 诊断词与标准词含有一对多的情况',
    'CHIP-CTC': '临床试验筛选标准短文本分类, https://tianchi.aliyun.com/dataset/dataDetail?dataId=95414#1',
    'CHIP-STS': '平安医疗科技疾病问答迁移学习, https://tianchi.aliyun.com/dataset/dataDetail?dataId=95414#1',
    'IMCS': '该数据集收集了真实的在线医患对话，并进行了多层次（Multi-Level）的人工标注，包含命名实体、对话意图、症状标签、医疗报告等，并在CCL 2021会议上举办了第一届智能对话诊疗评测比赛 （http://www.fudan-disc.com/sharedtask/imcs21/index.html）',
    'EMR_NER-PSE': '基于ccks2019医疗电子病历实体识别数据集制作的prompt-span-extraction数据集',
    'DuEE-Fin': '金融领域篇章级别事件抽取数据集， 共包含13个已定义好的事件类型约束和1.15万中文篇章（存在部分非目标篇章作为负样例），其中6900训练集，1150验证集和3450测试集，'
}



def explain(key: str):

    """
    Explain the meaning of a key.

    :param key: The key to explain.
    :return: The explanation.
    """

    if key in explain_dict:
        return explain_dict[key]
    else:
        return 'No explanation for this key.'
    
def chinese(tokenizer: str = 'basic'):
    import spacy
    from nlhappy.components.tokenizer import all_tokenizers
    tokenizer = all_tokenizers[tokenizer]
    nlp = spacy.blank('zh')
    nlp.tokenizer = tokenizer(nlp.vocab)
    return nlp

def nlp():
    import spacy
    return spacy.blank('zh')


def load(path: str, config: dict = {}, disable: list=[]):
    """load spacy nlp pipeline from path

    Args:
        path (str): path to the pipeline
        disable(list): a list of pipe names to disable

    Returns:
        nlp: spacy nlp pipeline
    """
    import spacy
    nlp = spacy.load(name=path, disable=disable, config=config)
    return nlp


from spacy.lang.zh import Chinese
from ..models import BertCRF, BertTokenClassification
import shutil
from thinc.api import Config
import os
import torch
from spacy.tokens import Span

models = {
    'bert_crf': BertCRF, 
    'bert_token_classification': BertTokenClassification
    }

class Tokenclassifier:
    '''token分类spacy pipeline
    - ckpt: 模型保存路径
    '''
    def __init__(self, nlp, name:str, ckpt: str, model: str, device: str, sentence_level: bool):
        self.nlp = nlp
        self.pipe_name = name
        self.ckpt = ckpt
        self.sentence_level = sentence_level
        self.model_name = model
        self.device = torch.device(device)
        try:
            self.model = models[self.model_name].load_from_checkpoint(self.ckpt)
            self.model.to(self.device)
            self.model.freeze()
        except Exception:
            pass


        
    def __call__(self, doc):
        if self.sentence_level:
            all_spans = []
            for sent in doc.sents:
                text = ''.join([t.text for t in sent])
                spans = self.model.predict(text, self.device)
                for span in spans:
                    s = Span(doc, span[0]+sent.start, span[1]+1+sent.start, span[2])
                    all_spans.append(s)
            doc.set_ents(all_spans)
        else:
            all_spans = []
            text = ''.join([t.text for t in doc])
            spans = self.model.predict(text, self.device)
            for span in spans:
                s = Span(doc, span[0], span[1]+1, span[2])
                all_spans.append(s)
            doc.set_ents(all_spans)
        return doc
    
    def to_disk(self, path:str, exclude):
        # 复制原来模型参数到新的路径
        if not os.path.exists(path):
            os.mkdir(path)
        shutil.copy(self.ckpt, path)
        # 重写NLP配置文件config.cfg 改变pipeline的ckpt路径
        # nlp_path = str(path).split(self.pipe_name)[0]
        # config_path = os.path.join(nlp_path, 'config.cfg')
        # config = Config().from_disk(config_path)
        # # config['components'][self.pipe_name]['ckpt_name'] = ckpt_name
        # config.to_disk(config_path)

    def from_disk(self, path:str, exclude):
        nlp_path = str(path).split(self.pipe_name)[0]
        config_path = os.path.join(nlp_path, 'config.cfg')
        config = Config().from_disk(config_path)
        ckpt_name = config['components'][self.pipe_name]['ckpt'].split('/')[-1]
        ckpt_path = os.path.join(path, ckpt_name)
        self.model = models[self.model_name].load_from_checkpoint(ckpt_path)
        self.model.freeze()

@Chinese.factory('token_classifier',assigns=['doc.ents'],default_config={'model':'bert_crf', 'device':'cpu'})
def get_tokencat(nlp, name, ckpt, model, device, sentence_level):
    return Tokenclassifier(nlp, name, ckpt, model, device, sentence_level)


from spacy.lang.zh import Chinese
from ..models import BertTextClassification
import torch
from spacy.tokens import Doc
import os
import pickle
import logging

log = logging.getLogger()


Doc.set_extension('labels', default=[])

class BertTextClassifier:
    '''spacy 文本分类管道'''
    def __init__(self, 
                 nlp, 
                 name:str, 
                 device: str,
                 threshold: float):
        super().__init__()
        self.nlp = nlp
        self.pipe_name = name
        self.device = torch.device(device)
        self.threshold = threshold
        
    def __call__(self, doc: Doc) -> Doc:
        with torch.no_grad():
            scores = self.model.predict(doc.text, device=self.device)
            if scores[0][1] >= self.threshold:
                doc._.label = scores[0][0]
            doc.cats = dict(scores)
        return doc
    
    def init_model(self, model_or_path):
        if isinstance(model_or_path, BertTextClassification):
            self.model = model_or_path
            self.model.to(self.device)
            self.model.freeze()
            
        else:
            self.model= BertTextClassification.load_from_checkpoint(model_or_path)
            self.model.to(self.device)
            self.model.freeze()

    def to_disk(self, path:str, exclude):
        if not os.path.exists(path):
            os.mkdir(path=path)
        model = 'tc.pkl'
        model_path = os.path.join(path, model)
        with open(model_path, 'wb') as f:
            pickle.dump(self.model, f)
        
    def from_disk(self, path:str, exclude):
        model_path = os.path.join(path, 'tc.pkl')
        with open(model_path, 'rb') as f:
            self.model = pickle.load(f)
        try:
            self.model.to(self.device)
        except:
            log.info(f' to device {self.device} failed')
        self.model.freeze()

default_config = {'model': 'bert_tc', 'device':'cpu', 'threshold':0.8}
@Chinese.factory('text_classifier', default_config=default_config)
def get_text_classification(nlp, name:str, device:str, model: str, threshold: float):
    if model == 'bert_tc':
        return BertTextClassifier(nlp, name=name, device=device, threshold=threshold)


    